# Code from: https://www.jujens.eu/posts/en/2018/Jun/02/python-timeout-function/
import signal
import gc
from time import time
from contextlib import contextmanager


@contextmanager
def timeout(seconds):
    '''
    time in seconds
    '''
    # Register a function to raise a TimeoutError on the signal.
    signal.signal(signal.SIGALRM, raise_timeout)
    # Schedule the signal to be sent after ``time``.
    signal.setitimer(signal.ITIMER_REAL, seconds)

    try:
        yield
    except TimeoutError:
        raise TimeoutError
    finally:
        # Unregister the signal so it won't be triggered
        # if the timeout is not reached.
        signal.setitimer(signal.ITIMER_REAL, 0)
        signal.signal(signal.SIGALRM, signal.SIG_IGN)


def raise_timeout(signum, frame):
    raise TimeoutError


class Timer:
    def __init__(self):
        self.elapsed = None
        self.start = None

    def __enter__(self):
        gc.disable()
        self.start = time()

    def __exit__(self, exception_type, exception_value, traceback):
        self.elapsed = time() - self.start
        gc.enable()



import unittest
import numpy as np
import random
import string
from types import ModuleType
from pathlib import Path
from complexitest import estimator


order = {
    estimator.CONSTANT: 1,
    estimator.LOG: 2,
    estimator.LINEAR: 3,
    estimator.QUADRATIC: 4,
    estimator.CUBIC: 5,
    estimator.EXPONENTIAL: 6,
    estimator.FACTORIAL: 7,
}


class LeetCodeEstimatorTestCase(unittest.TestCase):
    def solutions(self, problem, extra_code=''):
        problem_dir = Path(__file__).parent / 'leet_code' / problem
        for filename in problem_dir.glob('*.py'):
            print(filename)
            with open(filename) as f:
                solution_name = filename.name
                expected_times = [t for t in solution_name.split('_') if not t.isnumeric()]
                _module = ModuleType('user_module')
                exec('from typing import *', _module.__dict__)
                exec('import collections', _module.__dict__)
                exec('import math', _module.__dict__)
                exec('import itertools', _module.__dict__)
                exec(extra_code, _module.__dict__)
                exec(f.read(), _module.__dict__)
                solution = _module.Solution()
                yield solution, expected_times, solution_name

    def test_two_sum(self):
        def gen(choose_i1i2):
            def g(n):
                numbers = np.random.randint(1000, size=n).tolist()
                i1, i2 = choose_i1i2(n)
                ts = numbers[i1] + numbers[i2]
                for i in range(n):
                    if i not in (i1, i2):
                        numbers[i] += ts + 1
                return numbers, ts
            return g

        for solution, expected_times, solution_name in self.solutions('two_sum'):
            complexity = estimator.estimate_complexity(
                solution.twoSum,
                (
                    gen(lambda n: (n - 2, n - 1)),
                    gen(lambda n: (0, 1)),
                ),
                min_n=10,
                samples=1000
            )
            self.assertEqual(complexity, expected_times[0], msg=f"Didn't work for {solution_name}")

    def test_zigzag(self):
        def gen(n):
            letters = string.ascii_uppercase
            s = ''.join(random.choice(letters) for i in range(n))
            num_rows = random.randrange(1, n)
            return s, num_rows

        for solution, expected_times, solution_name in self.solutions('zigzag'):
            complexity = estimator.estimate_complexity(
                solution.convert,
                gen,
                min_n=10,
                samples=1000
            )
            self.assertLessEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

    def test_trapping_rain_water(self):
        def gen_len(n):
            return (np.random.randint(100, size=n).tolist(), )

        heights = np.random.randint(10, size=100).tolist()
        def gen_high(n):
            return ([h * n for h in heights], )

        for solution, expected_times, solution_name in self.solutions('trapping_rain_water'):
            complexity = estimator.estimate_complexity(
                solution.trap,
                gen_len,
                min_n=10,
                samples=1000
            )
            self.assertEqual(complexity, expected_times[0], msg=f"Didn't work for {solution_name}")

            complexity = estimator.estimate_complexity(
                solution.trap,
                gen_high,
                min_n=1,
                samples=1000
            )
            self.assertEqual(complexity, expected_times[1], msg=f"Didn't work for {solution_name}")

    def test_all_possible_full_binary_trees(self):
        extra_code = '''
class TreeNode:
    def __init__(self, val=0, left=None, right=None):
        self.val = val
        self.left = left
        self.right = right
        '''

        for solution, expected_times, solution_name in self.solutions('all_possible_full_binary_trees', extra_code):
            complexity = estimator.estimate_complexity(
                solution.allPossibleFBT,
                lambda n: (n if n % 2 == 1 else n + 1, ),  # Only makes sense for odd numbers
                min_n=1,
                samples=1000
            )
            self.assertEqual(complexity, expected_times[0], msg=f"Didn't work for {solution_name}")

    def test_bitwise_ors_of_subarrays(self):
        for solution, expected_times, solution_name in self.solutions('bitwise_ors_of_subarrays'):
            complexity = estimator.estimate_complexity(
                solution.subarrayBitwiseORs,
                lambda n: (np.random.randint(int(1e9), size=n).tolist(), ),
                min_n=1,
                samples=1000,
            )
            self.assertLessEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

    def test_kth_symbol_in_grammar(self):
        def gen_n(n):
            k = 1
            return n, k

        def gen_k(k):
            n = 15
            if k > 2**(n-1):
                k = 2**(n-1) - 1
            return n, k

        for solution, expected_times, solution_name in self.solutions('kth_symbol_in_grammar'):
            complexity = estimator.estimate_complexity(
                solution.kthGrammar,
                gen_n,
                min_n=1,
                samples=1000,
                max_non_poly=20,
            )
            self.assertLessEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

            complexity = estimator.estimate_complexity(
                solution.kthGrammar,
                gen_k,
                min_n=1,
                max_n=2**14,
                samples=1000,
            )
            self.assertLessEqual(order[complexity], order[expected_times[1]], msg=f"Didn't work for {solution_name}")

    def test_max_product_word_length(self):
        def gen_word(l):
            return ''.join(random.choice(string.ascii_lowercase) for _ in range(10))

        def gen_words(n):
            return [gen_word(10) for _ in range(n)],

        def gen_lengths(n):
            return [gen_word(n) for _ in range(10)],

        for solution, expected_times, solution_name in self.solutions('max_product_word_length'):
            complexity = estimator.estimate_complexity(
                solution.maxProduct,
                gen_words,
                min_n=2,
                samples=100,
            )
            self.assertLessEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

            complexity = estimator.estimate_complexity(
                solution.maxProduct,
                gen_lengths,
                min_n=2,
                samples=100,
            )
            self.assertLessEqual(order[complexity], order[expected_times[1]], msg=f"Didn't work for {solution_name}")

    def test_rotate_image(self):
        for solution, expected_times, solution_name in self.solutions('rotate_image'):
            complexity = estimator.estimate_complexity(
                solution.rotate,
                lambda n: (np.random.randint(256, size=(n, n)).tolist(), ),
                min_n=1,
                samples=100,
            )
            self.assertEqual(complexity, expected_times[0], msg=f"Didn't work for {solution_name}")

    def test_jump_game_ii(self):
        for solution, expected_times, solution_name in self.solutions('jump_game_ii'):
            complexity = estimator.estimate_complexity(
                solution.jump,
                lambda n: (np.random.randint(1, min(n, 5), size=n).tolist(), ),
                min_n=2,
                samples=500,
            )
            self.assertLessEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

    def test_first_missing_positive(self):
        def gen(n):
            l = list(range(1, n+1))
            random.shuffle(l)
            return l,

        for solution, expected_times, solution_name in self.solutions('first_missing_positive'):
            complexity = estimator.estimate_complexity(
                solution.firstMissingPositive,
                gen,
                min_n=1,
                samples=100,
            )
            self.assertEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")

    def test_valid_parentheses(self):
        def gen(n):
            if n % 2 == 1:
                n += 1
            op = n // 2
            cl = n // 2
            ret = ''
            while op or cl:
                if op >= cl:
                    op -= 1
                    ret += '('
                elif op > 0:
                    if random.choice((0, 1)):
                        op -= 1
                        ret += '('
                    else:
                        cl -= 1
                        ret += ')'
                else:
                    cl -= 1
                    ret += ')'
            return ret,

        for solution, expected_times, solution_name in self.solutions('valid_parentheses'):
            complexity = estimator.estimate_complexity(
                solution.isValid,
                gen,
                min_n=2,
                samples=100,
            )
            self.assertEqual(order[complexity], order[expected_times[0]], msg=f"Didn't work for {solution_name}")



import unittest
import numpy as np
import random
from complexitest import estimator


class EstimatorTestCase(unittest.TestCase):
    def test_constant(self):
        def do_nothing(n):
            return n

        complexity = estimator.estimate_complexity(
            do_nothing,
            lambda n: (n, ),
        )
        self.assertEqual(complexity, estimator.CONSTANT)

    def test_log_function(self):
        def add_log(l, start=0):
            if start == len(l) - 1:
                return 0
            return l[start] + add_log(l, (start + len(l)) // 2)

        complexity = estimator.estimate_complexity(
            add_log,
            lambda n: (np.random.randint(1000, size=n).tolist(), ))
        self.assertEqual(complexity, estimator.LOG)

    def test_linear_function(self):
        def add_all(l):
            s = 0
            for i in l:
                s += i
            return s

        complexity = estimator.estimate_complexity(
            add_all, lambda n: (np.random.randint(1000, size=n).tolist(), ))
        self.assertEqual(complexity, estimator.LINEAR)

    def test_quadratic_function(self):
        def add_all(l):
            s = 0
            for i in l:
                for j in l:
                    s += i + j
            return s

        complexity = estimator.estimate_complexity(
            add_all, lambda n: (np.random.randint(1000, size=n).tolist(), ))
        self.assertEqual(complexity, estimator.QUADRATIC)

    def test_cubic_function(self):
        def add_all(l):
            s = 0
            for i in l:
                for j in l:
                    for k in l:
                        s += i + j + k
            return s

        complexity = estimator.estimate_complexity(
            add_all, lambda n: (np.random.randint(1000, size=n).tolist(), ))
        self.assertEqual(complexity, estimator.CUBIC)

    def test_exponential_function(self):
        def rec(n, i=0):
            if i == n:
                return 1
            return rec(n, i + 1) + rec(n, i + 1)

        complexity = estimator.estimate_complexity(rec, lambda n: (n, ))
        self.assertEqual(complexity, estimator.EXPONENTIAL)

    def test_factorial_function(self):
        def permutations(l):
            if not l:
                return []
            ret = []
            for i, el in enumerate(l):
                ret.append([el] + permutations(l[:i] + l[i + 1:]))
            return ret

        complexity = estimator.estimate_complexity(
            permutations, lambda n: (np.random.randint(1000, size=n).tolist(), ))
        self.assertEqual(complexity, estimator.FACTORIAL)



import sys
import numpy as np
import itertools
import math
from .timer import timeout, Timer

CONSTANT = 'constant'  # Doesn't work... keeps mistaking it for log
LOG = 'logarithmic'
LINEAR = 'linear'
QUADRATIC = 'quadratic'
CUBIC = 'cubic'
EXPONENTIAL = 'exponential'
FACTORIAL = 'factorial'

NON_POLYNOMIAL = [
    EXPONENTIAL,
    FACTORIAL,
]

COMPLEXITY_CLASSES = {
    CONSTANT: lambda n: [1],
    LOG: lambda n: [math.log(n, 2), 1],
    LINEAR: lambda n: [n, 1],
    QUADRATIC: lambda n: [n**2, 1],
    CUBIC: lambda n: [n**3, 1],
    EXPONENTIAL: lambda n: [2**n, 1],
    FACTORIAL: lambda n: [math.factorial(n), 1],
}


def _test_complexities(times, is_non_poly=False):
    complexities = {}
    for complexity, func in COMPLEXITY_CLASSES.items():
        if (not is_non_poly and complexity in NON_POLYNOMIAL) or \
           (is_non_poly and complexity not in NON_POLYNOMIAL):
            continue
        a, b = zip(*[[func(n), t] for n, t in times.items()])
        try:
            x, residuals, _, _ = np.linalg.lstsq(a, b, rcond=None)
            if residuals.size > 0:
                complexities[complexity] = x, residuals
        except TypeError:
            pass
    return complexities


class NGenerator:
    def __init__(self, min_n=1, max_n=None, max_non_poly=30, samples=100):
        self.min_n = min_n
        self.max_n = max_n
        self.is_non_poly = False
        self.max_non_poly = max_non_poly
        self.samples = samples
        self._timed_out = False
        self.n = 0

    def generate(self):
        count = 0
        used = set()
        largest_n = self.min_n
        searching_largest = False

        start = self.min_n
        self.n = None
        self._timed_out = False
        while count < self.samples:
            if self.n is None:
                self.n = start
            elif self._timed_out:
                break
            else:
                largest_n = max(self.n, largest_n)
                self.n *= 2
            self._timed_out = False
            if self.n in used:
                break
            if self.max_n and self.n > self.max_n:
                largest_n = self.max_n
                break
            yield self.n
            used.add(self.n)
            count += 1

        largest_n = max(largest_n, self.max_non_poly)
        self.n = self.min_n
        while count < self.samples:
            if self.n >= largest_n or self._timed_out:
                self.n = self.min_n
            self._timed_out = False
            yield self.n
            count += 1
            self.n += 1

    def timed_out(self):
        if self.n < self.max_non_poly:
            self.is_non_poly = True
        self._timed_out = True


def estimate_complexity(f,
                        args_generator,
                        kwargs_generator=lambda n: {},
                        min_n=1,
                        max_n=None,
                        max_wait=0.01,
                        gen_max_wait=0.1,
                        samples=100,
                        max_non_poly=30,
                        show_plot=False,
                        ret_pts=False):
    """
    Estimates the complexity for function f

    Args:
        f: Target function
        args_generator: Function that returns the arguments for a given input size n or tuple of Functions (in this case we choose the worst time)
        kwargs_generator: Function that returns the arguments for a given input size n or tuple of Functions (in this case we choose the worst time)
        min_n: Starting value for n
        max_n: Maximum value for n (ignored if None)
        max_wait: Timeout duration
        gen_max_wait: Timeout for arguments generation
        samples: Number of samples to use for complexity estimation
        max_non_poly: Maximum number to consider factorial and exponential as possible solutions
        show_plot: Whether to show plot or not
        ret_pts: Whether to also return points
    """
    def make_return(ret, pts=None):
        if ret_pts:
            return ret, pts
        return ret

    if not isinstance(args_generator, tuple):
        args_generator = (args_generator, )
    if not isinstance(kwargs_generator, tuple):
        kwargs_generator = (kwargs_generator, )
    times = {}
    gen = NGenerator(min_n=min_n, max_n=max_n, samples=samples, max_non_poly=max_non_poly)
    timer = Timer()
    for n in gen.generate():
        if n > 2147483647:  # Some arbitrary big number (maxint)
            return make_return(CONSTANT)
        alt_times = []
        timed_out = False
        for args_g, kwargs_g in itertools.product(args_generator, kwargs_generator):
            try:
                with timeout(gen_max_wait):
                    args = args_g(n)
                with timeout(gen_max_wait):
                    kwargs = kwargs_g(n)
                with timeout(max_wait), timer:
                    f(*args, **kwargs)
                alt_times.append(timer.elapsed)
                if timer.elapsed > max_wait:  # This may occur when the code captures the TimeoutError in a try except.
                    gen.timed_out()
                    timed_out = True
            except (TimeoutError, RecursionError):
                gen.timed_out()
                timed_out = True
        if alt_times and not timed_out:
            new_time = max(alt_times)
            if n not in times or times[n] < new_time:
                times[n] = new_time
    complexities = _test_complexities(times, gen.is_non_poly)
    ret, (ret_x, _) = min(complexities.items(), key=lambda v: v[1][1])
    pts = [(k, v, COMPLEXITY_CLASSES[ret](k) @ ret_x) for k, v in times.items()]
    if show_plot:
        import matplotlib.pyplot as plt

        plt.plot(list(times.keys()), list(times.values()))
        for complexity, (x, _) in complexities.items():
            func = COMPLEXITY_CLASSES[complexity]
            plt.plot(list(times.keys()), [func(n) @ x for n in times])
        plt.show()
    return make_return(ret, pts)


if __name__ == "__main__":
    from random import randint

    def f(l):
        total = 0
        for i in l:
            for j in l:
                total += i + j
        return total

    complexity = estimate_complexity(f,
                                     lambda n:
                                     ([randint(0, 1000) for _ in range(n)], ),
                                     show_plot=True)
    print(complexity)



#!/usr/bin/env python

import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='complexitest',
    version='0.0.1',
    description='Algorithm Complexity Estimator',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Andrew Kurauchi',
    author_email='andrew.kurauchi@gmail.com',
    url='https://github.com/toshikurauchi/complexity_estimator',
    packages=['complexitest'],
    install_requires=['numpy'],
    license="MIT",
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)
